!> @file user_lpm_init.f90
!--------------------------------------------------------------------------------!
! This file is part of PALM.
!
! PALM is free software: you can redistribute it and/or modify it under the terms
! of the GNU General Public License as published by the Free Software Foundation,
! either version 3 of the License, or (at your option) any later version.
!
! PALM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! PALM. If not, see <http://www.gnu.org/licenses/>.
!
! Copyright 1997-2016 Leibniz Universitaet Hannover
!--------------------------------------------------------------------------------!
!
! Current revisions:
! -----------------
! 
! 
! Former revisions:
! -----------------
! $Id: user_lpm_init.f90 1818 2016-04-06 15:53:27Z maronga $
!
! 1682 2015-10-07 23:56:08Z knoop
! Code annotations made doxygen readable 
! 
! 1416 2014-06-04 16:04:03Z suehring
! load module indices
! 
! 1359 2014-04-11 17:15:14Z hoffmann
! New particle structure integrated. 
!
! 1320 2014-03-20 08:40:49Z raasch
! kind-parameters added to all INTEGER and REAL declaration statements, 
! kinds are defined in new module kinds, 
! old module precision_kind is removed,
! comment fields (!:) to be used for variable explanations added to
! all variable declaration statements 
!
! 1036 2012-10-22 13:43:42Z raasch
! code put under GPL (PALM 3.9)
!
! 849 2012-03-15 10:35:09Z raasch
! routine renamed: user_init_particles -> user_lpm_init
!
! 211 2008-11-11 04:46:24Z raasch
! Former file user_interface.f90 split into one file per subroutine
!
! Description:
! ------------
!> Modification of initial particles by the user.
!------------------------------------------------------------------------------!
 SUBROUTINE user_lpm_init ()
 
    USE indices

    USE kinds
    
    USE particle_attributes
    
    USE user

    IMPLICIT NONE

    INTEGER(iwp) ::  ip   !<
    INTEGER(iwp) ::  jp   !< 
    INTEGER(iwp) ::  kp   !< 
    INTEGER(iwp) ::  n    !< 

!+- MONA:
!-- Read the location of particle sources from a file and initialize the lpm model
    if (read_source_areas) then
       call location_message('Call lpm_create_particle_read',.true.)
    endif
!+- MONA end

 END SUBROUTINE user_lpm_init

!+- MONA:
!------------------------------------------------------------------------------!
! Description:
! ------------
!> Read particle source areas from a file PARTICLE_SOURCE. 
!> Needs: source_areas = .T. in &userpar
!------------------------------------------------------------------------------!
 SUBROUTINE user_read_lpm_source ( phase )

    USE arrays_3d,                                                             &
        ONLY: zw

    USE control_parameters,                                                    &
        ONLY: dz, io_blocks, io_group, message_string, coupling_char

    USE grid_variables,                                                        &
        ONLY:  ddx, dx, ddy, dy

    USE indices,                                                               &
        ONLY:  nx, nxl, nxlg, nxrg, nxr, ny, nyn, nys, nyng, nysg, nz, nzb,    &
               nzb_w_inner, nzt

    USE kinds

    USE lpm_exchange_horiz_mod,                                                &
        ONLY: lpm_exchange_horiz, lpm_move_particle, realloc_particles_array

    USE lpm_pack_arrays_mod,                                                   &
        ONLY: lpm_pack_all_arrays

    USE particle_attributes,                                                   &
        ONLY:   alloc_factor, curvature_solution_effects, grid_particles,      &
                initial_weighting_factor, iran_part,                           &
                maximum_number_of_particles, min_nr_particle,                  &
                number_of_particles, number_of_particle_groups,                &
                offset_ocean_nzt, particles, particle_groups, particle_type,   &
                particles_per_point, pdx, pdy, pdz, prt_count, psb, psl, psn,  &
                psr, pss, pst, random_start_position, seed_follows_topography, &
                total_number_of_particles, zero_particle

    USE pegrid

    USE pmc_interface,                                                         &
        ONLY:  cpl_id

    USE random_function_mod,                                                   &
        ONLY:  random_function

!+- MONA:
    USE user,                                                                  &
        ONLY: source_areas

    IMPLICIT  NONE

    INTEGER(iwp)               ::  alloc_size  !<
    INTEGER(iwp)               ::  i           !<
    INTEGER(iwp)               ::  ip          !<
    integer(iwp)               ::  ii          !<    
    INTEGER(iwp)               ::  j           !<
    INTEGER(iwp)               ::  jp          !<
    INTEGER(iwp)               ::  kp          !<
    INTEGER(iwp)               ::  loop_stride !<
    INTEGER(iwp)               ::  n           !<
    INTEGER(iwp)               ::  new_size    !<

    INTEGER(iwp), INTENT(IN)   ::  phase       !<

    INTEGER(iwp), DIMENSION(nzb:nzt+1,nysg:nyng,nxlg:nxrg) ::  local_count !<
    INTEGER(iwp), DIMENSION(nzb:nzt+1,nysg:nyng,nxlg:nxrg) ::  local_start !<

    LOGICAL                    ::  first_stride !<

    REAL(wp)                   ::  pos_x !<
    REAL(wp)                   ::  pos_y !<
    REAL(wp)                   ::  pos_z !<

    TYPE(particle_type),TARGET ::  tmp_particle !<

    if ( phase == 1 )  then
  
       allocate(source_areas(0:ny,0:nx))
       source_areas(:,:) = 0

       do ii = 0, io_blocks-1
          if (ii == io_group) then
             open(210, file='PARTICLE_SOURCE'//TRIM(coupling_char), &
                status='OLD', form='FORMATTED', err=106)

             do j = ny, 0, -1
                read(210, *, err=107, end=107) ( source_areas(j,i), i = 0,nx )
             enddo

             goto 108

106             message_string = 'file PARTICLE_SOURCE' //TRIM(coupling_char) // ' does not exist'
                CALL message( 'lpm_create_particle_read', 'UI0007', 1, 2, 0, 6, 0 )

107             message_string = 'errors in file PARTICLE_SOURCE' //TRIM(coupling_char)
                CALL message( 'lpm_create_particle_read', 'UI0007', 1, 2, 0, 6, 0 )

108             close(210) 

          endif
#if defined( __parallel )
             CALL MPI_BARRIER( comm2d, ierr )
#endif
       enddo
    endif


!
!-- Calculate particle positions and store particle attributes, if
!-- particle is situated on this PE
    DO  loop_stride = 1, 2
       first_stride = (loop_stride == 1)
       IF ( first_stride )   THEN
          local_count(:,:,:) = 0           ! count number of particles
       ELSE
          local_count(:,:,:) = prt_count(:,:,:)   ! Start address of new particles
       ENDIF

       n = 0
       DO  i = 1, number_of_particle_groups
          
          pos_z = psb(i)
          pos_y = 0

          DO WHILE ( pos_y <= ny*dy )

             IF ( pos_y >= ( nys - 0.5_wp ) * dy  .AND.  &
                  pos_y <  ( nyn + 0.5_wp ) * dy )  THEN

                pos_x = 0

         xloop: DO WHILE ( pos_x <= nx*dx )

                   IF ( pos_x >= ( nxl - 0.5_wp ) * dx  .AND.  &
                           pos_x <  ( nxr + 0.5_wp ) * dx )  THEN

!
!--                   Determine the grid indices of the particle position
                      ip = ( pos_x + 0.5_wp * dx ) * ddx
                      jp = ( pos_y + 0.5_wp * dy ) * ddy
                      kp = pos_z / dz + 1 + offset_ocean_nzt
                          
                      if (source_areas(jp,ip) == i )  then

          !               write(400+myid,*) jp, ip, i, source_areas(jp,ip) 

                         DO  j = 1, particles_per_point

                            n = n + 1
                            tmp_particle%x             = pos_x
                            tmp_particle%y             = pos_y
                            tmp_particle%z             = pos_z
                            tmp_particle%age           = 0.0_wp
                            tmp_particle%age_m         = 0.0_wp
                            tmp_particle%dt_sum        = 0.0_wp
                            tmp_particle%dvrp_psize    = 0.0_wp !unused
                            tmp_particle%e_m           = 0.0_wp
                            IF ( curvature_solution_effects )  THEN
!
!--                            Initial values (internal timesteps, derivative)
!--                            for Rosenbrock method
                               tmp_particle%rvar1      = 1.0E-12_wp
                               tmp_particle%rvar2      = 1.0E-3_wp
                               tmp_particle%rvar3      = -9999999.9_wp
                            ELSE
!
!--                            Initial values for SGS velocities
                               tmp_particle%rvar1      = 0.0_wp
                               tmp_particle%rvar2      = 0.0_wp
                               tmp_particle%rvar3      = 0.0_wp
                            ENDIF
                            tmp_particle%speed_x       = 0.0_wp
                            tmp_particle%speed_y       = 0.0_wp
                            tmp_particle%speed_z       = 0.0_wp
                            tmp_particle%origin_x      = pos_x
                            tmp_particle%origin_y      = pos_y
                            tmp_particle%origin_z      = pos_z
                            tmp_particle%radius        = particle_groups(i)%radius
!                            tmp_particle%weight_factor = initial_weighting_factor

!+- By Fabian Hoffman:
!--                         Usually, weight_factor is set by the parameter
!--                         initial_weighting_factor. Unfortunately, there is 
!--                         only one parameter for all particle groups.
!--                         For investigating different sources of pollution
!--                         of different strength, I would suggest to set the 
!--                         weighting factor for each group individually, e.g.:
                            tmp_particle%weight_factor = 2**(i-1) !* 100.0_wp


                            tmp_particle%class         = 1
                            tmp_particle%group         = i
                            tmp_particle%tailpoints    = 0     !unused
                            tmp_particle%particle_mask = .TRUE.
                            tmp_particle%tail_id       = 0     !unused

                            IF ( seed_follows_topography )  THEN
!
!--                            Particle height is given relative to topography
                               kp = kp + nzb_w_inner(jp,ip)
                               tmp_particle%z = tmp_particle%z + zw(kp)
                               IF ( kp > nzt )  THEN
                                  pos_x = pos_x + pdx(i)
                                  CYCLE xloop
                               ENDIF
                            ENDIF

                            local_count(kp,jp,ip) = local_count(kp,jp,ip) + 1
                            IF ( .NOT. first_stride )  THEN
                               IF ( ip < nxl  .OR.  jp < nys  .OR.  kp < nzb+1 )  THEN
                                  write(6,*) 'xl ',ip,jp,kp,nxl,nys,nzb+1
                               ENDIF
                               IF ( ip > nxr  .OR.  jp > nyn  .OR.  kp > nzt )  THEN
                                  write(6,*) 'xu ',ip,jp,kp,nxr,nyn,nzt
                               ENDIF
                               grid_particles(kp,jp,ip)%particles(local_count(kp,jp,ip)) = tmp_particle
                            ENDIF

                         ENDDO

                      endif

                   ENDIF

                   pos_x = pos_x + pdx(i)

                ENDDO xloop

             ENDIF

             pos_y = pos_y + pdy(i)

          ENDDO

       ENDDO

       IF ( first_stride )  THEN
          DO  ip = nxl, nxr
             DO  jp = nys, nyn
                DO  kp = nzb+1, nzt
                   IF ( phase == 1 )  THEN
                      IF ( local_count(kp,jp,ip) > 0 )  THEN
                         alloc_size = MAX( INT( local_count(kp,jp,ip) *        &
                            ( 1.0_wp + alloc_factor / 100.0_wp ) ),            &
                            min_nr_particle )
                      ELSE
                         alloc_size = min_nr_particle
                      ENDIF
                      ALLOCATE(grid_particles(kp,jp,ip)%particles(1:alloc_size))
                      DO  n = 1, alloc_size
                         grid_particles(kp,jp,ip)%particles(n) = zero_particle
                      ENDDO
                   ELSEIF ( phase == 2 )  THEN
                      IF ( local_count(kp,jp,ip) > 0 )  THEN
                         new_size   = local_count(kp,jp,ip) + prt_count(kp,jp,ip)
                         alloc_size = MAX( INT( new_size * ( 1.0_wp +          &
                            alloc_factor / 100.0_wp ) ), min_nr_particle )
                         IF( alloc_size > SIZE( grid_particles(kp,jp,ip)%particles) )  THEN
                           CALL realloc_particles_array(ip,jp,kp,alloc_size)
                         ENDIF
                      ENDIF
                   ENDIF
                ENDDO
             ENDDO
          ENDDO
       ENDIF
    ENDDO

    local_start = prt_count+1
    prt_count   = local_count
!
!-- Add random fluctuation to particle positions
    IF ( random_start_position )  THEN
       DO  ip = nxl, nxr
          DO  jp = nys, nyn
             DO  kp = nzb+1, nzt
                number_of_particles = prt_count(kp,jp,ip)
                IF ( number_of_particles <= 0 )  CYCLE
                particles => grid_particles(kp,jp,ip)%particles(1:number_of_particles)

                DO  n = local_start(kp,jp,ip), number_of_particles              !Move only new particles
                   IF ( psl(particles(n)%group) /= psr(particles(n)%group) )  THEN
                      particles(n)%x = particles(n)%x +                        &
                                   ( random_function( iran_part ) - 0.5_wp ) * &
                                   pdx(particles(n)%group)
                   ENDIF
                   IF ( pss(particles(n)%group) /= psn(particles(n)%group) )  THEN
                      particles(n)%y = particles(n)%y +                        &
                                   ( random_function( iran_part ) - 0.5_wp ) * &
                                   pdy(particles(n)%group)
                   ENDIF
                   IF ( psb(particles(n)%group) /= pst(particles(n)%group) )  THEN
                      particles(n)%z = particles(n)%z +                        &
                                   ( random_function( iran_part ) - 0.5_wp ) * &
                                   pdz(particles(n)%group)
                   ENDIF
                ENDDO
!
!--             Identify particles located outside the model domain
                CALL lpm_boundary_conds( 'bottom/top' )
             ENDDO
          ENDDO
       ENDDO
!
!--    Exchange particles between grid cells and processors
       CALL lpm_move_particle
       CALL lpm_exchange_horiz

    ENDIF
!
!-- In case of random_start_position, delete particles identified by 
!-- lpm_exchange_horiz and lpm_boundary_conds. Then sort particles into blocks, 
!-- which is needed for a fast interpolation of the LES fields on the particle 
!-- position.
    CALL lpm_pack_all_arrays

!
!-- Determine maximum number of particles (i.e., all possible particles that 
!-- have been allocated) and the current number of particles
    DO  ip = nxl, nxr
       DO  jp = nys, nyn
          DO  kp = nzb+1, nzt
             maximum_number_of_particles = maximum_number_of_particles         &
                                           + SIZE(grid_particles(kp,jp,ip)%particles)
             number_of_particles         = number_of_particles                 &
                                           + prt_count(kp,jp,ip)
          ENDDO
       ENDDO
    ENDDO
!
!-- Calculate the number of particles of the total domain
#if defined( __parallel )
    IF ( collective_wait )  CALL MPI_BARRIER( comm2d, ierr )
    CALL MPI_ALLREDUCE( number_of_particles, total_number_of_particles, 1, &
    MPI_INTEGER, MPI_SUM, comm2d, ierr )
#else
    total_number_of_particles = number_of_particles
#endif

    RETURN

 END SUBROUTINE user_read_lpm_source

!+- MONA end

